from argparse import Namespace
from typing import Callable, Any, List
from dataclasses import dataclass, field
import abc


@dataclass
class Parameter:
    """Class representing argparse parameter"""

    name: str
    help: str
    action: str = None


@dataclass
class Command:
    __metaclass__ = abc.ABCMeta

    """Class reprensenting argparse subparser"""
    name: str
    help: str
    params: List[Parameter] = field(default_factory=list)

    @abc.abstractmethod
    def execute(self, args: Namespace = None) -> None:
        """Execute the command

        Args:
            args (Namespace, optional): List of args provided by argparse. Defaults to None.
        """
        raise NotImplementedError()


@dataclass
class Group:
    id: int
    name: str
    path: str
    description: str
    visibility: str
    web_url: str
    full_path: str
    parent_id: int


class Project:
    def __init__(
        self,
        _id: int,
        description: str,
        web_url: str,
        name: str,
        path: str,
        path_with_namespace: str,
        star_count: int,
        open_issues_count: int,
        namespace_id: int,
        namespace_name: str,
        ssh_url_to_repo: str,
    ) -> None:
        self.id = _id
        self.description = description
        self.web_url = web_url
        self.name = name
        self.path_with_namespace = path_with_namespace
        self.path = path
        self.star_count = star_count
        self.open_issues_count = open_issues_count
        self.namespace = {"id": namespace_id, "name": namespace_name}
        self.ssh_url_to_repo = ssh_url_to_repo

    def __str__(self) -> str:
        return f'Porject({self.id}, "{self.path_with_namespace}")'

    def __repr__(self) -> str:
        return self.__str__()


@dataclass
class ProjectQuestion:
    message: str
    name: str
    default: str
    help: str
    validate: Callable[[str], bool] = None
    transform: Callable[[str], bool] = None
    non_api: bool = False
    response: Any = field(init=False)

    def __post_init__(self):
        if self.message == "" or self.name == "":
            raise ValueError("No value provided for message or name")

    def transform_response(self, response: Any) -> None:
        if response is None:
            return

        if self.transform is not None:
            self.response = self.transform(response)

    def set_response(self, response: Any) -> None:
        if response is None:
            self.response = self.default

        self.response = response



#!/usr/bin/python
"""hmp - HEPIA project manager

This script is a helper for all students at HEPIA
to assist the creation of a new coding project with
a gitlab (https://gitedu.hesge.ch) repo directly link
to the local instance.

@author : Tanguy Cavagna <tanguy.cavagna@etu.hesge.ch>
@copyrigth : Copyright 2021, hpm
"""

from . import cli_args as arguments
from .commands.constants import COMMANDS


def main():
    args = arguments.parse_args()

    if args.command is not None:
        for command in COMMANDS:
            if args.command == command.name:
                command.execute(args)


if __name__ == "__main__":
    main()



import requests
import json
import pathlib
import yaml
from urllib.parse import urlencode
from typing import List, Union
from requests.models import Response
from rich import print
from rich.console import Console
from yaml.loader import SafeLoader
from ._typing import Group, Project
from .errors import APIError


token: str
base_uri: str
console = Console()

# =============
# Decorator
# =============
def need_api_token(func):
    """Decorator for call with needed token

    Args:
        func (Callable): Function to call
    """

    def wrapper_func(*args, **kwargs):
        global token, base_uri

        # =============
        # Retrieve config
        # =============
        try:
            with open(
                pathlib.Path(__file__).resolve().parent.absolute() / "config.yaml", "r"
            ) as f:
                config = dict(yaml.load(f, Loader=SafeLoader))

                if not config["token"]:
                    raise Exception("No token found")

                token = config["token"]
                base_uri = config["target"]  # Always have a default value
        except (FileNotFoundError, Exception) as e:
            if isinstance(e, FileNotFoundError):
                print(
                    "Before making any calls, you need to setup at least your token by doing : hpm setup --token <TOKEN_VALUE>"
                )
                print(
                    "You can also setup a custom target uri by adding the parameter `--target <TARGET_URI>`"
                )
                exit()

            token = None

        if not token:
            print(
                "No token as been register. Please do so using `hpm setup --token <TOKEN_VALUE>.`"
            )
            exit()  # Force program exit because nothing will run correctly if no token as been provided
        return func(*args, **kwargs)

    return wrapper_func


# =============
# Utility function
# =============
def _create_group_from_json(json_response: dict) -> Group:
    """Create a group from a json response

    Args:
        json_response (dict): Response from api

    Returns:
        Group: Newly created group
    """
    return Group(
        json_response["id"],
        json_response["name"],
        json_response["path"],
        json_response["description"],
        json_response["visibility"],
        json_response["web_url"],
        json_response["full_path"],
        json_response["parent_id"],
    )


def _create_project_from_json(json_response: dict) -> Project:
    """Create a project from a json response

    Args:
        json_response (dict): Response from api

    Returns:
        Project: Newly created project
    """
    return Project(
        json_response["id"],
        json_response["description"],
        json_response["web_url"],
        json_response["name"],
        json_response["path"],
        json_response["path_with_namespace"],
        json_response["star_count"],
        json_response["open_issues_count"],
        json_response["namespace"]["id"],
        json_response["namespace"]["name"],
        json_response["ssh_url_to_repo"],
    )


def _validate_api_response(response: Response) -> bool:
    """Validate an api call response

    Args:
        response (Response): Provided response

    Returns:
        bool: Response validation status
    """
    if not response.ok:
        # Custom message for most known error status codes
        if response.status_code == 404:
            print("Cannot reach API endpoint", response.status_code)
        elif response.status_code == 500:
            print("Server-side error", response.status_code)
        else:
            print("Error during API call", response.status_code)

        return False

    return True


# =============
# API calls
# =============
@need_api_token
def get_owned_groups() -> List[Group]:
    """Get user owned groups

    Returns:
        Group[]: List of owned groups
    """
    url = f"{base_uri}/groups?owned=true&per_page=100"

    payload = {}
    headers = {"Private-Token": token}

    with console.status("[bold green]💾 Getting groups..."):
        response = requests.request("GET", url, headers=headers, data=payload)

    if not _validate_api_response(response):
        return None

    response = json.loads(response.text)

    groups: List[Group] = []
    # @see: https://gitedu.hesge.ch/help/api/groups.md#list-groups
    for json_group in response:
        groups.append(_create_group_from_json(json_group))

    return groups


@need_api_token
def get_owned_projects(page=1):
    """Get user owned projects

    Args:
        page (int): Pagination number

    Return:
        Project[]: List of owned projects
    """
    url = f"{base_uri}/projects?membership=true&page={page}"

    payload = {}
    headers = {"Private-Token": token}

    with console.status("[bold green]💾 Getting projects..."):
        response = requests.request("GET", url, headers=headers, data=payload)

    if not _validate_api_response(response):
        return None

    response_json = json.loads(response.text)

    p_list: List[Project] = []
    # @see: https://gitedu.hesge.ch/help/api/projects.md#list-all-projects
    for json_project in response_json:
        p_list.append(_create_project_from_json(json_project))

    return {
        # @see: https://stackoverflow.com/a/45087988 for informations on headers
        "headers": {
            "X-Next-Page": response.headers.get("X-Next-Page"),
            "X-Prev-Page": response.headers.get("X-Prev-Page"),
            "X-Total-Pages": response.headers.get("X-Total-Pages"),
        },
        # Sort the project by full namespace for easier print later
        "list": sorted(p_list, key=lambda k: k.path_with_namespace),
    }


@need_api_token
def create_project(params: list) -> Union[Project, APIError]:
    """Create a new project

    Args:
        params (list): Param list

    Returns:
        Project: Newly created project
    """
    url = f"{base_uri}/projects?"

    # Build url parameters
    url_params = {}
    for param in params:
        if param["non_api"]:
            continue
        
        if param["value"] is not None and param["value"] != "":
            url_params[param["key"]] = param["value"]

        # Special case for namespace check
        if param["key"] == "namespace_id" and param["value"] == -1:
            return APIError(message="Namespace not found")

    url += urlencode(url_params)

    payload = {}
    headers = {"Private-Token": token}

    with console.status("[bold green]🚀 Creating project..."):
        response = requests.request("POST", url, headers=headers, data=payload)

    if not _validate_api_response(response):
        return None

    response = json.loads(response.text)

    return _create_project_from_json(response)


@need_api_token
def get_subgroups(parent: Group) -> List[Group]:
    """Get the parent subgroups list

    Args:
        parent (Group): Parent group

    Returns:
        List[Group]: List of subgroups
    """
    url = f"{base_uri}/groups/{parent.id}/subgroups"

    payload = {}
    headers = {"Private-Token": token}

    with console.status("[bold green]💾 Getting subgroups...") as status:
        response = requests.request("GET", url, headers=headers, data=payload)

    if not _validate_api_response(response):
        return None

    response = json.loads(response.text)

    groups: List[Group] = []
    for json_group in response:
        groups.append(_create_group_from_json(json_group))

    return groups


@need_api_token
def get_group_projects(group: Group) -> List[Project]:
    """Get the provided group's project list

    Args:
        group (Group): Group to fetch from

    Returns:
        List[Project]: List of projects
    """
    url = f"{base_uri}/groups/{group.id}/projects"

    payload = {}
    headers = {"Private-Token": token}

    with console.status("[bold green]💾 Getting group's projects..."):
        response = requests.request("GET", url, headers=headers, data=payload)

    if not _validate_api_response(response):
        return None

    response_json = json.loads(response.text)

    p_list: List[Project] = []
    for json_project in response_json:
        p_list.append(_create_project_from_json(json_project))

    return p_list


@need_api_token
def get_group(group_id: int) -> Group:
    """Get group details

    Args:
        group_id (int): Wanted group

    Returns:
        Group: Found group
    """
    url = f"{base_uri}/groups/{group_id}"

    payload = {}
    headers = {"Private-Token": token}

    with console.status("[bold green]💾 Getting group..."):
        response = requests.request("GET", url, headers=headers, data=payload)

    if not _validate_api_response(response):
        return None

    response = json.loads(response.text)

    return _create_group_from_json(response)



import os
import platform
import sys
from datetime import datetime
from rich import print
from rich.console import Console
from rich.panel import Panel
from rich.columns import Columns
from rich.text import Text

CURSOR_UP_ONE = "\x1b[1A"
ERASE_LINE = "\x1b[2K"


def print_command_header(command: str):
    """Global header

    Args:
        command (str): Executed command
    """
    print(
        Panel(
            Columns(
                [
                    Text("Made By Tanguy Cavagna with 💖", style="italic"),
                    Text(
                        datetime.now().strftime("%Y-%m-%d, %H:%M:%S"),
                        justify="right",
                        style="italic",
                    ),
                ],
                expand=True,
                title="[bold bright_white]HES-SO[/bold bright_white] - [bold red1]HEPIA[/bold red1]",
            ),
            border_style="bright_white",
            title=f"Command: [orange_red1]{command}[/orange_red1]",
            title_align="left",
        )
    )


def print_ssh_key_warning_message():
    """Warning message for user regarding ssh key"""
    have_ssh = None
    console = Console()
    prompt = "⚠️ [bright_white]Do you have an ssh key setup[/bright_white] [dim italic](y, n)[/dim italic] "

    while have_ssh not in ["y", "n"]:
        have_ssh = console.input(prompt)
        del_last_line()

    print(f"{prompt}[cyan]{have_ssh}[/cyan]")

    have_ssh = True if have_ssh == "y" else False

    # Exit execution if no setup
    if not have_ssh:
        print()
        print("Please setup an ssh key before using the `create-app` command.")
        help_link = "[italic bold cyan][link=https://docs.gitlab.com/ee/ssh/#generate-an-ssh-key-pair]generate ssh key[/link][/italic bold cyan]"
        print(f"> Refer to {help_link} to correctly setup one.")
        print()
        exit()


def clr_scr():
    """Clear screen on all os"""
    if platform.system().lower() == "windows":
        os.system("cls")
    else:
        os.system("clear")


def del_last_line():
    """Delete the last console line"""
    sys.stdout.write(CURSOR_UP_ONE)
    sys.stdout.write(ERASE_LINE)



from typing import List
from ._typing import Group, Project
from .api import get_owned_groups


def insert_subnamspace(parent: dict, subnamespace: Group):
    """Insert a new subnamespace in the parent

    Args:
        parent (dict): Parent namespace
        subnamespace (Group): Child namespace

    Returns:
        dict: Updated parent
    """
    if parent["id"] == subnamespace.parent_id:
        parent["children"].append(
            {
                "id": subnamespace.id,
                "name": subnamespace.name,
                "full_path": subnamespace.full_path,
                "url": subnamespace.web_url,
                "children": [],
                "projects": [],
            }
        )
        return parent

    for _next_namespace in parent["children"]:
        if isinstance(_next_namespace, dict):
            item = insert_subnamspace(_next_namespace, subnamespace)
            if item is not None:
                return item


def insert_project_in_namespace(namespaces: dict, project: Project):
    """Insert a project in the corresponding namespace

    Args:
        namespaces (dict): Root namespaces
        project (Project): Project to be inserted

    Returns:
        dict: Updated namespaces dict
    """
    if namespaces["id"] == project.namespace["id"]:
        namespaces["projects"].append(project)
        return namespaces

    for _next_namespace in namespaces["children"]:
        _next = insert_project_in_namespace(_next_namespace, project)
        if _next is not None:
            return _next


def create_namespace_tree(groups: List[Group], projects: List[Project]) -> list:
    """Create a namespace array containing every projects

    Args:
        groups (List[Group]): Group list
        projects (List[Project]): Project list

    Returns:
        list: List of all namespaces.\ne.g.:\n[\n{\n\t'id': 1234,\n\t'name': 'my-group',\n\t'children': [{...}],\n\t'projects': [<Project>,...]\n\n}, {...}\n]
    """
    parent = []
    for group in groups:
        if group.parent_id is not None:
            for p in parent:
                p = insert_subnamspace(p, group)
        else:
            parent.append(
                {
                    "id": group.id,
                    "name": group.name,
                    "full_path": group.full_path,
                    "url": group.web_url,
                    "children": [],
                    "projects": [],
                }
            )

    tmp = projects.copy()
    for project in projects:
        for par in parent:
            if insert_project_in_namespace(par, project) is not None:
                tmp.remove(project)
                break

    # Now projects holds only personal projects (not the ones in groups)
    projects = tmp
    if len(projects) > 0:
        parent.append(
            {
                "id": projects[0].namespace["id"],
                "name": "Personal",
                "url": f'https://gitedu.hesge.ch/users/{projects[0].namespace["name"]}/projects',
                "children": [],
                "projects": [],
            }
        )
    for project in projects:
        insert_project_in_namespace(parent[-1], project)

    return parent


def get_namespace_id(namespace_name: str) -> int:
    """Get the namespace id to privide to the api

    Args:
        namespace_name (str): Namespace name

    Returns:
        int: Namespace id
    """
    GROUPS = get_owned_groups()

    if GROUPS is None:
        return -1

    _id = [group.id for group in GROUPS if group.full_path == namespace_name]

    if len(_id) > 0:
        return _id[0]
    else:
        return -1



import argparse
from .commands.constants import COMMANDS


def parse_args():
    parser = argparse.ArgumentParser(description="HEPIA project manager.")

    subparsers = parser.add_subparsers(help="Custom commands", dest="command")

    for c in COMMANDS:
        tmp_parser = subparsers.add_parser(c.name, help=c.help)

        if c.params != []:
            for p in c.params:
                if p.action is not None:
                    tmp_parser.add_argument(
                        p.name,
                        help=p.help,
                        action=p.action,
                    )
                    continue

                tmp_parser.add_argument(
                    p.name,
                    help=p.help,
                )

    args = parser.parse_args()

    args_values = vars(args)
    if not any(args_values.values()):
        parser.error("No commands provided.")

    return args
